#!/usr/libexec/platform-python
import re
import time
import os

__author__ = 'intigua'

import http.client
import subprocess

import intiguautil
from intiguautil import Ssh

NGINX_CONF_TEMPLATE = """
upstream backend8080  {{
  server {0}:443;
}}

upstream backend9090 {{
  server {0}:443;
}}
"""

MONITOR_INTERVAL_SECS = 30
DR_ACTIVE_FILE = "dr_active"
DRY_RUN = False

class IntiguaInstance(object):
    def __init__(self, remote):
        self.remote = remote

    def get_conn(self):
        return http.client.HTTPConnection(self.remote, 8080)


    def get_response(self, conn):
        conn.request("GET", "/vmanage-server")
        return conn.getresponse()

    def ping(self):
        conn = self.get_conn()
        return self.get_response(conn).status < 400

    def check_health(self):
        try:
            return self.ping()
        except Exception:
            return False

class DisasterRecovery(object):

    def __init__(self, mainInstance, drInstnace, standby):
        self.mainInstnace = mainInstance
        self.drInstnace = drInstnace
        self.standby = standby

    def recover(self):
        "it is assumed that the backup instance contains update to date data."
        # stop first instance if possible:
        stop_intigua(self.mainInstnace.remote)

        stop_sync(self.drInstnace.remote)

        # start the second instance
        start_intigua(self.drInstnace.remote)

        # check that the new instance is up
        while not self.drInstnace.check_health():
            time.sleep(1)
            # TODO need to add timeout

        if not self.standby:
            # reconfigure nginx
            self.redirect_to(self.drInstnace.remote)
        elif self.standby:
            # create 'dr_active' file for checking dr status
            open(DR_ACTIVE_FILE, 'a').close()

    def monitor(self, prog = None):
        while self.mainInstnace.check_health():
            time.sleep(MONITOR_INTERVAL_SECS)
        self.recover()
        if prog:
            subprocess.Popen(prog, shell=True)

    def redirect_to(self, remote):
        write_nginx_config(remote)

def stop_intigua(remote):
    print('Stopping intigua at ', remote)
    if DRY_RUN:
        return
    services = intiguautil.check_install_profile(remote)
    for service_name in services:
        stop_service(remote,service_name)


def start_intigua(remote):
    print('Starting intigua at ', remote)
    if DRY_RUN:
        return
    services = intiguautil.check_install_profile(remote)
    for service_name in services:
        start_service(remote,service_name)

def stop_sync(remote):
    stop_service(remote,"intigua-sync")

def start_sync(remote):
    start_service(remote,"intigua-sync")

def start_service(remote, service):
    print('Starting ',service, 'at', remote)
    if DRY_RUN:
        return
    Ssh(remote).run(["chkconfig", "--add", service])
    Ssh(remote).run(["service", service, "start"])

def stop_service(remote, service):
    print('Stopping ',service, 'at', remote)
    if DRY_RUN:
        return
    Ssh(remote).run(["service", service, "stop"])
    Ssh(remote).run(["chkconfig", "--del", service])


def write_nginx_config(remote):
        print('Redirecting nginx to', remote)
        if DRY_RUN:
            return
        conf = NGINX_CONF_TEMPLATE.format(remote)
        with open("/etc/nginx/conf.d/lb_server", "w") as f:
            f.write(conf)
        subprocess.call(["service", "nginx", "reload"])


def read_nginx_config():
        with open("/etc/nginx/conf.d/lb_server", "r") as f:
            conf = f.read()
        servers =  list(set(re.findall("server (.+):", conf)))
        if not servers:
            return None
        if len(servers) != 1:
            raise Exception("Configuration is not as expected")

        return servers[0]

def parse_args():
    import argparse   # only supported on python 2.7, rest of this script may be used with python 2.6
    parser = argparse.ArgumentParser(description='Intigua sync client')

    # Do not use the default paramter in the params that are also read from the config file!!
    parser.add_argument('--master', dest="master", type=str, help='Where to backup from. leave empty for config')
    parser.add_argument('--slave', dest="slave", type=str, help='Where to backup to. leave empty for config')
    parser.add_argument('--test',dest="test", action='store_true', help='just test, dont really change anything.')
    parser.add_argument('--dr-hook',dest="prog", type=str, help='Program to run on DR event.')
    parser.add_argument('--init',dest="init", action='store_true', help='Init master and slave (i.e start tomcat in them. use with --standby if running on slave.')
    parser.add_argument('--monitor',dest="monitor", action='store_true', help='Monitor DR changes.')
    parser.add_argument('--conf', dest='conf', type=str, default ="/etc/intigua/dr.conf",
                   help='Configuration file, to avoid settings the parameters in the command line every time. Parameters given on the command line override the configutration file. Default location is /etc/intigua/dr.conf')
    parser.add_argument('--createconfig',dest="configfile", action='store_true', help='Create the configuration at a location. does nothing else.')
    parser.add_argument('--pidfile', dest='pidfile', type=str, default=None, help='write a pid file and daemonize')
    parser.add_argument('--standby', dest='standby', action='store_true', help='Standby Mode. Use the standby server to monitor instead of a dedicated nginx server, Use this with the --init and --monitor.')
    return parser.parse_args()

def start(args):

    from configparser import ConfigParser
    config = ConfigParser()
    config.read(args.conf)

    if args.configfile:
        if not config.has_section("Servers"):
            config.add_section("Servers")
        if args.master:
            config.set("Servers","master", args.master)
        else:
            config.remove_option("Servers", "master")
        if args.slave:
            config.set("Servers", "slave", args.slave)
        else:
            config.remove_option("Servers", "slave")
        if args.prog:
            config.set("Servers", "dr-hook", args.prog)
        else:
            config.remove_option("Servers", "dr-hook")

        with open(args.conf, 'w') as configfile:
            config.write(configfile)
        return

    get_or_none = lambda section, configParam: config.get(section, configParam) if config.has_option(section, configParam) else None
    if not args.master:
        args.master = get_or_none("Servers","master")
    if not args.slave:
        args.slave = get_or_none("Servers","slave")
    if not args.prog:
        args.prog = get_or_none("Servers","dr-hook")


    master_host = args.master
    slave_host = args.slave

    if args.test:
        global DRY_RUN
        DRY_RUN = True

    curHost = None
    if not args.standby:
        curHost = read_nginx_config()

    if args.init:
        if not args.standby:
            write_nginx_config(master_host)
        elif os.path.isfile(DR_ACTIVE_FILE):
            os.remove(DR_ACTIVE_FILE) # clear 'dr_active' file
        start_intigua(master_host)
        stop_intigua(slave_host)
        start_sync(slave_host)

    # check if already in dr mode
    if args.standby and os.path.isfile(DR_ACTIVE_FILE):
        raise Exception('Already in DR mode. nothing to do.')
    elif curHost == slave_host:
        raise Exception('Already in DR mode. nothing to do.')

    # test ssh
    if not DRY_RUN:
        intiguautil.test_ssh(master_host)
        intiguautil.test_ssh(slave_host)

    if args.monitor:
        dr = DisasterRecovery(IntiguaInstance(master_host), IntiguaInstance(slave_host), args.standby)
        dr.monitor(args.prog)


def main():
    args = parse_args()

    if args.pidfile:
        from daemonize import Daemonize
        daemon = Daemonize(app="intigua-dr-monitor", pid=args.pidfile, action=lambda : start(args))
        daemon.start()
    else:
        start(args)

if __name__ == "__main__":
    main()
